/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail;

import java.io.Serializable;

/**
 * <p>
 * A data class for holding url and the title of a page.
 * </p>
 *
 * <p>
 * This class will be created by the NodeDiscovery implementations (when discoverying the current node)
 * and by the PathDiscovery implementations to return the path from the current node to the root node.
 * Each node in that path will use this class(The path is composed of NodePair instances while NodePair
 * instance owns instance of this class).
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is immutable and thread safe.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 1.0
 *
 */
public class BreadCrumbNode implements Serializable {
    /**
     * <p>
     * Represents the title associated with this node.
     * </p>
     *
     * <p>
     * This variable is set by the constructor and is immutable.
     * This value of this can be null if the title should default to the title found in the PathDiscovery.
     * This value should not be null if it's a node that is part of a path returned by the PathDiscovery implementation.
     * This variable is referenced by the associated getter.
     * </p>
     *
     * @since 1.0
     */
    private final String title;

    /**
     * <p>
     * Represents the url (including query string) associated with this node.
     * This variable is set by the constructor, is immutable and will never be null.
     * This variable is referenced by the associated getter.
     * </p>
     *
     * @since 1.0
     */
    private final String url;

    /**
     * <p>
     * Constructor for BreadCrumbNode with title and url specified.
     * </p>
     *
     * <p>
     * <b>Note: </b> If title is an empty (trim'd) string, null should be used instead.
     * </p>
     *
     * @param url A non-null, non-empty (trim'd) url of page represented by this node
     * @param title A possibly null, possibly empty (trim'd) title of page represented by this node
     *
     * @throws IllegalArgumentException if url is null or an empty (trim'd) string
     *
     * @since 2.0
     */
    public BreadCrumbNode(String url, String title) {
        Util.checkString(url, "url");

        this.url = url;
        this.title = (Util.isEmptyString(title) ? null : title);
    }

    /**
     * <p>
     * Simply the getter method for the title of the page.
     * </p>
     *
     * @return A possibly null, non-empty (trim'd) string representing the title of the page
     *
     * @since 1.0
     */
    public String getTitle() {
        return title;
    }

    /**
     * <p>
     * Simply the getter method for the Url. The returned string will be the URL with any query string.
     * </p>
     *
     * @return A non-null, non-empty (trim'd) string representing the url (including query string) of the page
     *
     * @since 1.0
     */
    public String getUrl() {
        return url;
    }
}
