/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.topcoder.util.objectfactory.InvalidClassSpecificationException;
import com.topcoder.util.objectfactory.ObjectFactory;
import com.topcoder.util.objectfactory.impl.ConfigManagerSpecificationFactory;
import com.topcoder.util.objectfactory.impl.IllegalReferenceException;
import com.topcoder.util.objectfactory.impl.SpecificationConfigurationException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.PathDiscovery;
import com.topcoder.web.ui.tag.breadcrumbtrail.Util;

import javax.servlet.jsp.PageContext;

/**
 * <p>
 * This implementation of the PathDiscovery will attempt to create a static site map and then apply dynamic
 * information to that site map. The dynamic information can either be in matching nodes with different query
 * strings or in new nodes that will be assumed to map to the last known node.
 * </p>
 *
 * <p>
 * This class can be used by the BreadCrumbTrailTag to discover the path from the given node to the root node.
 * This class will use session level attributes in page context to store the current path that is known and the
 * current site map.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 *
 */
public class DynamicPathDiscovery implements PathDiscovery {
    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the current known path.
     * </p>
     *
     * <p>
     * This constant will be used by this class to retrieve the current path. The current path will be of NodePair[]
     * type representing the known path from the root (0th index) to the prior page (length-1th index).
     * </p>
     *
     */
    public static final String CURRENT_PATH = DynamicPathDiscovery.class.getName() + ".currentPath";

    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the site map.
     * </p>
     *
     * <p>
     * This constant will be used by this class to retrieve the site map. The site map will be a Map with
     * NodeMatcher as the key and a List of NodeMatcher as the value and can be empty. This map represents
     * the NodeMatcher parent to it's children NodeMatcher(s) if any. The list can be empty if there are no
     * children of the parent (i.e. a leaf node)
     * </p>
     *
     */
    public static final String CURRENT_SITEMAP = DynamicPathDiscovery.class.getName() + ".siteMap";

    /**
     * <p>
     * Constructs the DynamicPathDiscovery object using no constructor parameters.
     * </p>
     *
     */
    public DynamicPathDiscovery() {
        // empty
    }

    /**
     * <p>
     * This method attempts to determine the path from the current node to the root and return that path.
     * </p>
     *
     * <p>
     * Before performing the shortest path searching in site map, two quick steps take first:
     * <ol>
     * <li>Iterate the current path backwards and see if the node matches the NodeMatcher for any element
     * in current path</li>
     * <li>Take the last element in the path and see if the current node matches any of the children of
     * that node.</li>
     * </ol>
     * </p>
     *
     * <p>
     * Note, this method will return null when no shortest path has been found from the siteMap and no currentPath
     * existed indicating we have no idea what the path is.
     * </p>
     *
     * @param currentNode A non-null current node
     * @param pageContext A non-null page context
     * @return A null if the path can't be found or a non-empty array representing the path from root node (0th index)
     * to current node (size-1th index)
     *
     * @throws IllegalArgumentException if either arg is null
     * @throws BreadCrumbException if an exception occurred while processing
     */
    public BreadCrumbNode[] getPath(BreadCrumbNode currentNode, PageContext pageContext) throws BreadCrumbException {
        Util.checkNull(currentNode, "currentNode");
        Util.checkNull(pageContext, "pageContext");

        NodePair[] newPath = null;

        // discovery the current path
        NodePair[] currentPath = getCurrentPath(pageContext);

        // discovery the site map
        Object siteObj = pageContext.getAttribute(CURRENT_SITEMAP, PageContext.SESSION_SCOPE);
        Map siteMap = null;
        if (siteObj instanceof Map) {
            siteMap = (Map) siteObj;
        } else {
            currentPath = null;
            siteMap = createSiteDiscovery().getSiteMap();
        }

        // searches the current path to see if the current node(of BreadCrumbNode type) matches any node
        // (of NodePair type) in the path
        if (currentPath != null) {
            newPath = searchCurrentPath(currentNode, currentPath);
            if (newPath != null) {
                return updatePath(newPath, currentNode, siteMap, pageContext);
            }
        }

        // searches the last node to see if the current node is a child of it
        if (currentPath != null && siteMap != null) {
            newPath = searchLastNode(currentNode, currentPath, siteMap);
            if (newPath != null) {
                return updatePath(newPath, currentNode, siteMap, pageContext);
            }
        }

        // searches the site map for shortest path
        if (siteMap != null) {
            // for detail algorithm, refer to Util#shortestPath method
            List candicatePathList = Util.shortestPath(siteMap, currentNode);

            if (candicatePathList != null && !candicatePathList.isEmpty()) {
                // select the most commonality path with the current path
                newPath = mostMatchesPath(currentPath, candicatePathList);
                return updatePath(newPath, currentNode, siteMap, pageContext);
            }
        }

        // no shortest path has been found from the siteMap and no currentPath existed
        // set the current path to null indicating we have no idea what the path is
        // otherwise the current node is appended to the current path
        if (currentPath != null) {
            newPath = appendCurNodeToCurPath(currentNode, currentPath);
        } else {
            newPath = null;
        }

        return updatePath(newPath, currentNode, siteMap, pageContext);
    }

    /**
     * <p>
     * Searches the last node if it has a non-null nodeMathcer instance attached on the current path to see whether the
     * current node is any children of that node.
     * If yes, return new path, otherwise return null.
     * </p>
     *
     * <p>
     * The new path is consisted of the current path and the current node.
     * </p>
     *
     * @param currentNode the current node that represents the current page
     * @param currentPath the current path that represents the history pages visited by user
     * @param siteMap the site map
     *
     * @return the new path if the current node is a child of the last node in current path, otherwise return null.
     */
    private NodePair[] searchLastNode(BreadCrumbNode currentNode, NodePair[] currentPath, Map siteMap) {

        // get the last node in the current path
        NodeMatcher nodeMatcher = currentPath[currentPath.length - 1].getNodeMatcher();

        // if it doesn't have non-null nodeMatcher instance, then null is returned
        if (nodeMatcher == null) {
            return null;
        }

        List children = (List) siteMap.get(nodeMatcher);

        if (children == null) {
            return null;
        }

        // iterate the children, if one of them matches the current node, then new path is generated and returned.
        for (Iterator it = children.iterator(); it.hasNext();) {
            NodeMatcher childNodeMatcher = (NodeMatcher) it.next();
            if (childNodeMatcher.matches(currentNode)) {
                // the current path and the current node will make the new current path
                NodePair[] newPath = new NodePair[currentPath.length + 1];
                System.arraycopy(currentPath, 0, newPath, 0, currentPath.length);
                newPath[currentPath.length] = new NodePair(currentNode, childNodeMatcher);
                return newPath;
            }
        }

        // no children of last node matches current node, null is returned.
        return null;

    }

    /**
     * <p>
     * Searches the current path backwards (size-1 to 0 index) to see if one of them matches the current node,
     * if yes, then new path is returned, otherwise null is returned.
     * </p>
     *
     * <p>
     * The new path is from the root node matcher in current path to the node matcher before the node matcher that the
     * current node(of BreadCrumbNode type) matches plusing the current node.
     * </p>
     *
     * @param currentNode the current node represents the current page
     * @param currentPath the current path visited by user
     *
     * @return the new path if the current node matches any of the nodes on the current path
     */
    private NodePair[] searchCurrentPath(BreadCrumbNode currentNode, NodePair[] currentPath) {
        NodePair[] newPath = null;
        // iterate the current path backwards(size-1 to 0 index)
        for (int i = currentPath.length - 1; i >= 0; i--) {
            if (nodeMatches(currentPath[i], currentNode)) {
                if (i != currentPath.length - 1) {
                    newPath = new NodePair[i + 1];
                    System.arraycopy(currentPath, 0, newPath, 0, i);
                    newPath[i] = new NodePair(currentNode, currentPath[i].getNodeMatcher());
                } else {
                    newPath = currentPath;
                }
            }
        }
        return newPath;
    }

    /**
     * <p>
     * Gets the current path from page context, if found, then return it, otherwise return null.
     * </p>
     *
     * @param pageContext A non-null page context
     *
     * @return the current path in page context, if not found, null will be returned
     */
    private NodePair[] getCurrentPath(PageContext pageContext) {
        Object pathObj = pageContext.getAttribute(CURRENT_PATH, PageContext.SESSION_SCOPE);
        NodePair[] currentPath = null;
        if (pathObj instanceof NodePair[]) {
            currentPath = (NodePair[]) pathObj;
        }
        return currentPath;
    }

    /**
     * <p>
     * Returns the path that has the most commonality with the current path starting with the root node forward in
     * the given candicatePathList.
     * </p>
     *
     * @param currentPath the current path that visited by user
     * @param candicatePathList that all the shortest path from any root node to the current node
     *
     * @return the path that has the most commonality with the current path starting with the root node forward in
     * the given candicatePathList.
     */
    private NodePair[] mostMatchesPath(NodePair[] currentPath, List candicatePathList) {
        int commonNodes = 0;
        int shortestPathIndex = 0;

        // iterate the current path to identify the index of the path has the most commonality with the current path
        // the common node number is calculated too.
        if (currentPath != null) {
            for (int curIndex = 0; curIndex < candicatePathList.size(); curIndex++) {
                List candicatePath = (List) candicatePathList.get(curIndex);
                int curCommonNodes = getCommonNodesNumber(currentPath, candicatePath);
                if (curCommonNodes > commonNodes) {
                    commonNodes = curCommonNodes;
                    shortestPathIndex = curIndex;
                }
            }
        }

        List path = (List) candicatePathList.get(shortestPathIndex);
        NodePair[] newPath = new NodePair[path.size()];

        // copy the common nodes from the current path
        if (currentPath != null && commonNodes != 0) {
            System.arraycopy(currentPath, 0, newPath, 0, commonNodes);
        }

        // get the new nodes from the shortest path
        for (int i = commonNodes; i < path.size(); i++) {
            NodeMatcher nodeMatcher = (NodeMatcher) path.get(i);
            newPath[i] = new NodePair(new BreadCrumbNode(nodeMatcher.getUrl(), nodeMatcher.getTitle()), nodeMatcher);
        }
        return newPath;
    }

    /**
     * <p>
     * Returns the common node number between the current path and the given path starting with the root node forward.
     * </p>
     *
     * @param currentPath the current path that visited by user
     * @param candicatePath one of the shortest path found in site map
     *
     * @return the common node number between the current path and the given path starting with the root node forward.
     */
    private int getCommonNodesNumber(NodePair[] currentPath, List candicatePath) {
        int curCommonNodes = 0;
        // compares the elements from 0-index
        for (int i = 0; i < candicatePath.size(); i++) {
            NodeMatcher nodeMatcher = (NodeMatcher) candicatePath.get(i);
            if (i < currentPath.length && nodeMatcher.equals(currentPath[i].getNodeMatcher())) {
                curCommonNodes++;
                continue;
            } else {
                break;
            }
        }
        return curCommonNodes;
    }

    /**
     * <p>
     * Appends the current node to the current path and return the new path.
     * </p>
     *
     * @param curNode the current node represents the current page
     * @param curPath the current path that visited by user
     *
     * @return the new path by appending the current node to the current path
     */
    private NodePair[] appendCurNodeToCurPath(BreadCrumbNode curNode, NodePair[] curPath) {
        NodePair[] newPath = null;
        // if curPath is null, then only
        if (curPath != null) {
            newPath = new NodePair[curPath.length + 1];
            System.arraycopy(curPath, 0, newPath, 0, curPath.length);
            newPath[curPath.length] = new NodePair(curNode, null);
        } else {
            newPath = new NodePair[1];
            newPath[0] = new NodePair(curNode, null);
        }
        return newPath;
    }

    /**
     * <p>
     * Returns true if the node matches the nodePair, otherwise return false.
     * </p>
     *
     * <p>
     * The match is done via:
     * <ol>
     * <li>If the nodeMatcher in nodePair is null then node urls must be equal in order to match.</li>
     * <li>If the nodeMatcher in nodePair is non-null then the node must match the nodeMatcher.</li>
     * </ol>
     * </p>
     *
     * @param nodePair NodePair instance represents one element on the current path
     * @param node BreadCrumbNode instance represents the current page
     *
     * @return true if the given node matches the nodePair, otherwise return false
     */
    private boolean nodeMatches(NodePair nodePair, BreadCrumbNode node) {
        if (nodePair.getNodeMatcher() == null && nodePair.getNode().getUrl().equals(node.getUrl())) {
            return true;
        }
        if (nodePair.getNodeMatcher() != null && nodePair.getNodeMatcher().matches(node)) {
            return true;
        }
        return false;
    }

    /**
     * <p>
     * Updates the CURRENT_PATH and CURRENT_SITEMAP attributes in page context in session scope and converts the path
     * to suitable form and returns it.
     * </p>
     *
     * @param path the new path calculated
     * @param currentNode the current node represents the current page
     * @param siteMap the site map
     * @param pageContext A non-null page context
     *
     * @return A null if the path can't be found or a non-empty array representing the path from root node (0th index)
     * to current node (size-1th index)
     */
    private BreadCrumbNode[] updatePath(NodePair[] path, BreadCrumbNode currentNode, Map siteMap,
            PageContext pageContext) {
        // CURRENT_PATH and CURRENT_SITEMAP updated
        pageContext.setAttribute(CURRENT_PATH, path, PageContext.SESSION_SCOPE);
        pageContext.setAttribute(CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);

        if (path == null || path.length == 0) {
            return null;
        }

        BreadCrumbNode[] nodes = new BreadCrumbNode[path.length];
        for (int i = 0; i < path.length; i++) {
            // The last element in the path should be set to currentNode instead
            if (i == path.length - 1) {
                if (currentNode.getTitle() == null && path[i].getNodeMatcher() != null) {
                    nodes[i] = new BreadCrumbNode(currentNode.getUrl(), path[i].getNodeMatcher().getTitle());
                } else {
                    nodes[i] = currentNode;
                }
            } else {
                nodes[i] = path[i].getNode();
            }
        }
        return nodes;
    }

    /**
     * <p>
     * Constructs a SiteDiscovery instance via object factory.
     * </p>
     *
     * <p>
     * The namespace to be used is the fully qualified name of DynamicPathDiscovery.
     * </p>
     *
     * @return a new NodeDiscovery instance dynamically created via object factory.
     * @throws BreadCrumbException when failed to create SiteDiscovery instance via object factory
     */
    private SiteDiscovery createSiteDiscovery() throws BreadCrumbException {
        try {
            ObjectFactory factory = new ObjectFactory(new ConfigManagerSpecificationFactory(
                    DynamicPathDiscovery.class.getName()));
            // in the config file, the property name should be the fully qualified name of SiteDiscovery interface
            // with dot be replaced with slash
            return (SiteDiscovery) factory.createObject(SiteDiscovery.class);
        } catch (SpecificationConfigurationException e) {
            throw new BreadCrumbException("SpecificationConfigurationException occurs, caused by " + e.getMessage(), e);
        } catch (IllegalReferenceException e) {
            throw new BreadCrumbException("IllegalReferenceException occurs, caused by " + e.getMessage(), e);
        } catch (InvalidClassSpecificationException e) {
            throw new BreadCrumbException("InvalidClassSpecificationException occurs, caused by " + e.getMessage(), e);
        }
    }
}
