/*
 * Copyright (C) 2006 TopCoder Inc., All Rights Reserved.
 */
package com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery;

import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbException;
import com.topcoder.web.ui.tag.breadcrumbtrail.BreadCrumbNode;
import com.topcoder.web.ui.tag.breadcrumbtrail.PathDiscovery;
import com.topcoder.web.ui.tag.breadcrumbtrail.Util;
import com.topcoder.web.ui.tag.breadcrumbtrail.pathdiscovery.nodematchers.RegexNodeMatcher;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.servlet.jsp.PageContext;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * <p>
 * This implementation of the PathDiscovery <b>and </b> SiteDiscovery will attempt to create a static site map from
 * either a file name or string source (both of which is an XML document).
 * </p>
 *
 * <p>
 * This class will parse the document to create a static map of Regex matchers(where each node can provide a regex
 * matching).
 * If used as a PathDiscovery, this provides a static view of the site (i.e. no dynamic capabilities).
 * </p>
 *
 * <p>
 * The BreadCrumbTrailTag will either construct this class directly (with a path override argument) or automatically
 * via the object factory.
 * </p>
 *
 * <p>
 * Note, this class will store the static site map to a session variable called CURRENT_SITEMAP so that the file/string
 * source needs to be parsed only a single time.
 * </p>
 *
 * <p>
 * <b>Thread Safety:</b> This class is thread safe in the context of the JSP rendering process.
 * </p>
 *
 * @author Pops, TCSDEVELOPER
 * @version 2.0
 * @since 2.0
 *
 */
public class XmlPathDiscovery implements SiteDiscovery, PathDiscovery {
    /**
     * <p>
     * This static constant represents the session scope attribute name that will contain the site map.
     * </p>
     *
     * <p>
     * This constant will be used by this class to retrieve the site map. The site map will be a Map with
     * NodeMatcher as the key and a List of NodeMatcher as the value and can be empty. This map represents
     * the NodeMatcher parent to it's children NodeMatcher(s) if any. The list can be empty if there are no
     * children of the parent (i.e. a leaf node).
     * </p>
     *
     */
    public static final String CURRENT_SITEMAP = XmlPathDiscovery.class.getName() + ".siteMap";

    /**
     * <P>
     * This static constant represents the bc_node tag name in site map xml file.
     * </p>
     *
     */
    private static final String BC_NODE = "bc_node".intern();

    /**
     * <P>
     * This static constant represents the title attribute name in site map xml file.
     * </p>
     *
     */
    private static final String TITLE = "title";

    /**
     * <P>
     * This static constant represents the url attribute name in site map xml file.
     * </p>
     *
     */
    private static final String URL = "url";

    /**
     * <P>
     * This static constant represents the pattern attribute name in site map xml file.
     * </p>
     *
     */
    private static final String PATTERN = "pattern";

    /**
     * <p>
     * Represents the source of the site data, it may be a raw xml string or a xml file path depends the
     * isFile flag.
     * </p>
     *
     * <p>
     * The source will be set by the constructor, is immutable and will never be null but may be an empty
     * (trim'd) string.
     * This variable is referencable by the getSiteMap method.
     * </p>
     *
     */
    private final String source;

    /**
     * <p>
     * Represents whether the source variable represents a file (true) or a raw xml string (false).
     * </p>
     *
     * <p>
     * The isFile will be set by the constructor and is immutable.
     * This variable is referencable by the getSiteMap method.
     * </p>
     *
     */
    private final boolean isFile;

    /**
     * <p>
     * Constructs the XmlPathDiscovery from the given filename.
     * </p>
     *
     * @param filename A non-null, non-empty (trim'd) string representing the file to use
     *
     * @throws IllegalArgumentException if the filename is a null or empty (trim'd) string
     */
    public XmlPathDiscovery(String filename) {
        this(filename, true);
    }

    /**
     * <p>
     * Constructs the XmlPathDiscovery from the given source and whether that source is a direct XML string
     * or a filename.
     * </p>
     *
     * @param source A non-null, non-empty (trim'd) string representing the source
     * @param isFile Whether the source is a file (true) or a string (false)
     *
     * @throws IllegalArgumentException if the source is null or an empty (trim'd) string
     */
    public XmlPathDiscovery(String source, boolean isFile) {
        Util.checkString(source, "source");

        this.source = source;
        this.isFile = isFile;
    }

    /**
     * <p>
     * Returns a map of NodeMatcher keys with a List of NodeMatcher elements as the value.
     * The key NodeMatcher represents the parent and the value (the list) represents the child values.
     * The value can be null or an empty list to represent no child elements (i.e. it's a leaf node).
     * </p>
     *
     * <p>
     * Besides the exception throw by SAX Parser, two structual errors will cause BreadCrumbException to be thrown:
     * <ol>
     * <li>When a NodeMatcher occurs more than once, at least two of them have child nodes</li>
     * <li>A NodeMathcer contains two equal child NodeMatchers.</li>
     * </ol>
     * </p>
     *
     * @return A non-null map of parent NodeMatcher to a list of NodeMatcher children
     *
     * @throws BreadCrumbException if parsing of data source failed
     */
    public Map getSiteMap() throws BreadCrumbException {
        // use the SAXParser bundled with JDK
        try {
            String validationFeature  = "http://xml.org/sax/features/validation";
            String schemaFeature = "http://apache.org/xml/features/validation/schema";

            SAXParserFactory factory = SAXParserFactory.newInstance();

            factory.setNamespaceAware(false);
            factory.setValidating(true);

            factory.setFeature(validationFeature, true);
            factory.setFeature(schemaFeature, true);

            SiteMapHandler handler = new SiteMapHandler();
            SAXParser parser = factory.newSAXParser();
            if (isFile) {
                // the file will be searched under relative path or class path.
                parser.parse(getSourceFile(source), handler);
            } else {
                parser.parse(new ByteArrayInputStream(source.getBytes()), handler);
            }
            return handler.getSiteMap();
        } catch (ParserConfigurationException e) {
            throw new BreadCrumbException("ParserConfigurationException occurs when parsing xml using SAX, caused by "
                    + e.getMessage(), e);
        } catch (SAXException e) {
            throw new BreadCrumbException(
                    "SAXException occurs when parsing xml using SAX, caused by " + e.getMessage(), e);
        } catch (IOException e) {
            throw new BreadCrumbException("IOException occurs when parsing xml using SAX, caused by " + e.getMessage(),
                    e);
        }
    }

    /**
     * <p>
     * Gets the input source file from the relative path, if it is not found, then try to find in the class path.
     * </p>
     *
     * <p>
     * If the input source file can not be found in relative path or class path, then the file in relative path will be
     * returned and the xml parser will throw exception expected.
     * </p>
     *
     * @param source the source file path to get the File instance
     *
     * @return a File instance of the given source path, searching in relative path and class path
     */
    private File getSourceFile(String source) {
        // find the source file in the relative file path.
        File sourceFile = new File(source);

        if (sourceFile.exists()) {
            return sourceFile;
        }

        // find the source file from class path
        URL url = XmlPathDiscovery.class.getClassLoader().getResource(source);

        if (url == null) {
            return sourceFile;
        } else {
            return new File(url.getFile());
        }
    }

    /**
     * <p>
     * This method attempts to determine the page from the current node to the root and return that path based on
     * the current site map.
     * </p>
     *
     * @param currentNode A non-null current node
     * @param pageContext A non-null page context
     *
     * @return A null if the path can't be found or a non-empty array representing the path from root node (0th index)
     * to current node (size-1th index)
     *
     * @throws IllegalArgumentException if any arg is null
     * @throws BreadCrumbException if an exception occurred while processing
     */
    public BreadCrumbNode[] getPath(BreadCrumbNode currentNode, PageContext pageContext) throws BreadCrumbException {
        Util.checkNull(currentNode, "currentNode");
        Util.checkNull(pageContext, "pageContext");

        // get the site map first
        Map siteMap = null;
        Object attrObj = pageContext.getAttribute(CURRENT_SITEMAP, PageContext.SESSION_SCOPE);
        if (attrObj instanceof Map) {
            siteMap = (Map) attrObj;
        } else {
            siteMap = getSiteMap();
            pageContext.setAttribute(CURRENT_SITEMAP, siteMap, PageContext.SESSION_SCOPE);
        }

        // if site map is null, then no path can be found
        if (siteMap == null) {
            return null;
        }

        // get the shortest path to a root node
        List pathList = Util.shortestPath(siteMap, currentNode);

        if (pathList == null || pathList.isEmpty()) {
            return null;
        } else {
            // get the first shortest path
            List path = (List) pathList.iterator().next();
            BreadCrumbNode[] nodes = new BreadCrumbNode[path.size()];

            // convert the NodePair[] instances to BreadCrumbNode instances in order to return
            for (int i = 0; i < path.size(); i++) {
                NodeMatcher nodeMatcher = (NodeMatcher) path.get(i);
                if (i == path.size() - 1) {
                    String title = currentNode.getTitle();
                    if (title == null) {
                        title = nodeMatcher.getTitle();
                    }
                    nodes[i] = new BreadCrumbNode(currentNode.getUrl(), title);
                } else {
                    nodes[i] = new BreadCrumbNode(nodeMatcher.getUrl(), nodeMatcher.getTitle());
                }
            }
            return nodes;
        }
    }

    /**
     * <p>
     * This class extends DefaultHandler and is used to parse xml content.
     * It is used by SAX parser and is used to parse site map xml.
     * </p>
     *
     * <p>
     * Note, SAX is used for loading xml file at the sake of speed and memory compared to DOM based parser.
     * </p>
     *
     * @author TCSDEVELOPER
     * @version 2.0
     *
     */
    private final class SiteMapHandler extends DefaultHandler {
        /**
         * <p>
         * A site map consists of NodeMatcher nodes.
         * It should have each unique NodeMatcher listed as the key and an List of NodeMatcher nodes that describe
         * the children of that key.
         * </p>
         *
         * <p>
         * It is
         * </p>
         */
        private Map siteMap;

        /**
         * <p>
         * A stack consists of NodeMatcher nodes.
         * when a new bc_node element starts, the corresponding NodeMatcher instance will be pushed in and when a
         * bc_node element ends, the corresponding NodeMatcher instance will be popped up.
         * </p>
         *
         * <p>
         * Note, the last element in bcNodeStack represents the parent node of the current processing node.
         * </p>
         */
        private LinkedList bcNodeStack;

        /**
         * <p>
         * This flag is set to true by a NodeMatcher node when it is expected to have no children, otherwise it's set
         * false (in endElement element).
         * </p>
         *
         * <p>
         * When processing the current nodes, if the childless is true, then exception will be thrown.
         * </p>
         */
        private boolean childless = false;

        /**
         * <p>
         * Creates a new SiteMapHandler instance.
         * </p>
         *
         * <p>
         * Note, it's package private.
         * </p>
         */
        SiteMapHandler() {
            siteMap = new HashMap();
            bcNodeStack = new LinkedList();
        }

        /**
         * <p>
         * If bc_node element starts, a NodeMatcher instance will created according to the attributes in the bc_node
         * element.
         * The parent-children relationship is set in siteMap in this method.
         * </p>
         *
         * <p>
         * There are two possible situations may cause exception:
         * <ol>
         * <li>When a NodeMatcher occurs more than once, at least two of them have child nodes</li>
         * <li>A NodeMathcer contains two equal child NodeMatchers.</li>
         * </ol>
         * </p>
         *
         * <p>
         * Note, when the SAX parser encounters a star-tag, this method is called.
         * </p>
         *
         * @param namespaceURI namespace uri
         * @param localName local name
         * @param qName qualified name
         * @param attr represents the attributes for the current element encountered
         *
         * @throws SAXException Any SAX exception.
         */
        public void startElement(String namespaceURI, String localName, String qName, Attributes attr)
            throws SAXException {
            if (qName.intern() == BC_NODE) {
                String title = attr.getValue(TITLE);
                String url = attr.getValue(URL);
                String pattern = attr.getValue(PATTERN);

                try {
                    NodeMatcher nodeMatcher = new RegexNodeMatcher(title, url, pattern);
                    List children = (List) siteMap.get(nodeMatcher);
                    if (children == null) {
                        siteMap.put(nodeMatcher, new LinkedList());
                    } else {
                        // the childless is true, then it must be set by direct parent node.
                        if (childless) {
                            throw new SAXException(
                                "When a NodeMatcher occurs more than once, at most one of them can have "
                                + "child nodes, others should be childless.");
                        }

                        // if the current NodeMathcer exists and has children, then the current bc_node element is just
                        // a reference and it is expected to have no children
                        if (!children.isEmpty()) {
                            childless = true;
                        }
                    }

                    // the last element in bcNodeStack represents the parent node of the current processing node.
                    if (!bcNodeStack.isEmpty()) {
                        NodeMatcher parentNodeMatcher = (NodeMatcher) bcNodeStack.getLast();
                        List siblings = (List) siteMap.get(parentNodeMatcher);
                        if (siblings.contains(nodeMatcher)) {
                            throw new SAXException("A NodeMathcer can't contain two equal child NodeMatchers.");
                        }
                        siblings.add(nodeMatcher);
                    }

                    // adds the current nodeMatcher to the stack, it will be removed in endElement method.
                    bcNodeStack.addLast(nodeMatcher);
                } catch (BreadCrumbException e) {
                    throw new SAXException(
                            "BreadCrumbException occurs when constructing RegexNodeMatcher instance, caused by "
                                    + e.getMessage(), e);
                }
            }
        }

        /**
         * <p>
         * If bc_node element ends, then last element in bcNodeStack will removed and childless is set to false to
         * identify the next bc_node element can have children.
         * </p>
         *
         * <p>
         * Note, when the SAX parser encounters a end-tag, this method is called.
         * </p>
         *
         * @param namespaceURI namespace uri
         * @param localName local name
         * @param qName qualified name
         *
         */
        public void endElement(String namespaceURI, String localName, String qName) {
            if (qName.intern() == BC_NODE) {
                bcNodeStack.removeLast();
                if (childless) {
                    childless = false;
                }
            }
        }

        /**
         * <p>
         * Returns the site map parsing from a xml file.
         * </p>
         *
         * <p>
         * In the returned map, key is an unique NodeMatcher instance and the value is a List of NodeMatcher nodes that
         * describe the children of that key.
         * </p>
         *
         * @return the site map parsing from a xml file.
         */
        Map getSiteMap() {
            return siteMap;
        }
    }
}
